# Decision Optimization for Endangered Species Recovery
# C. Ashton Drew
# (For new contact details see website: www.cashtondrew.com, ashton.drew@gmail.com)
# Affiliation for this grant-funded work:
# Biodiversity & Spatial Information Center
# NC Fish and Wildlife Cooperative Fish & Wildlife Research Unit
# Department of Applied Ecology
# North Carolina State University
# February 2015

# Although this program has been used by the U.S. Geological Survey (USGS), no warranty, expressed
# or implied, is made by the USGS or the U.S. Government as to the accuracy and functioning of the 
# program and related program material nor shall the fact of distribution constitute any such 
# warranty, and no responsibility is assumed by the USGS in connection therewith.

# This analysis takes Endangered Species Recovery Action data from an csv saved excel spreadsheet
# and identifies optimal decision sets given a budget limit, multivariate agency priorities, and
# knowledge uncertainty.

# Input data must meet content and formatting requirements outlined in 
# ES_Decisions_DataInputGuide.docx.  This code can be run with the example table 
# ES_Decisions_ExampleData.csv


# DEFINE BUDGET

# Manager inputs the amount of money available to allocate towards management actions
# for the upcoming FIVE YEAR PERIOD.  Managers best professional judgement.
# This is the amount left AFTER money is withdrawn for
#     1) big special project species (Manatee, Parrot, and Blackbird)

fiveyearbudget <- 500000
budget <- fiveyearbudget/5

#######################################################################

# IMPORT ELICITATION DATA

# Import decision input table as csv file.  Note: The Action column (recovery actions) MUST NOT HAVE COMMAS, 
# else the import will not work correctly.

setwd("E:/P_PriorityES/data/Post-Workshop/")

input <- read.csv("RecoveryActions_20150323.csv", header=TRUE, sep=",")

as.factor(input$Progress) # Change Progress field to factors as defined in data table
# 0 = not yet started, 1 = in progress, 2 = completed, 3 = discontinued, 4 = obsolete

summary(input) # Check input data
headers <- colnames(input) # Create vector of header names for future use

# SUBSET DATA INTO CANDIDATE and NON-CANDIDATE ACTIONS

# Rows where:
#   (1) Progress < 2 (In Progress or Not Yet Started), and 
#   (2) Prerequisites = N, and
#   (3) FWS share of annual costs are less then the annual budget.

# Calculate annual FWS portion of costs as: ((Cost-(Cost*Leverage))*1000)/Time where Cost are thousands of dollars 
# to complete the entire action including monitoring and reporting and Leverage is the percentage of costs 
# expected to be covered by partners.

input$FWSCostYr <- ((input$Cost-(input$Cost*input$Leverage))*1000)/input$Time

candidates <- input[ which(input$Prerequisite=='N'& input$Progress<2 & input$FWSCostYr<budget), ]
noncandidates <- input[ which(input$Prerequisite=='Y'| input$Progress>1 | input$FWSCostYr>budget), ]

write.table(noncandidates, "NonCandidateActions.csv", sep=",", row.names=FALSE)
write.table(candidates, "CandidateActions.csv", sep=",", row.names=FALSE)

########################################################################################

# CALCULATE PROBABILISTIC VALUES FOR UNCERTAIN ACTIONS GIVEN OBJECTIVES

# Uncertain variables in this model are:
#   (1) Successful Completion (CompX)
#   (2) Baseline Extinction (BaseX)
#   (3) With Action Extinction (ActX)
#   (4) No Action Risk (RiskX)
#   (5) Reclassification (ReclassX)
# In this model the Budget, Cost, Time, Benefits, and Leverage variables are treated as certain.

# SELECT HOW MANY SIMULATIONS (DECISION SETS) TO INCLUDE

sims = 1000

# Create table to receive simulation results

riskresults <- data.frame(matrix(NA, nrow=nrow(candidates), ncol=0))
riskresults[,1:3] <- candidates[,1:3] # Get Item, Species, and Action fields
persistresults <- data.frame(matrix(NA, nrow=nrow(candidates), ncol=0))
persistresults[,1:3] <- candidates[,1:3]
reclassresults <- data.frame(matrix(NA, nrow=nrow(candidates), ncol=0))
reclassresults[,1:3] <- candidates[,1:3]
balanceresults <- data.frame(matrix(NA, nrow=nrow(candidates), ncol=0))
balanceresults[,1:3] <- candidates[,1:3]

for (s in 1:sims){

  # DRAW EXPECTED PROBABILITY VALUES FROM PROBABILITY DISTRIBUTIONS
  
  # First calcuate a single draw for each probability given X and SD values.
  # If any result is greater than 1 or less than 0, set to 1 or 0, respectively.
  for(i in 1:nrow(candidates)){
    candidates$CompDrw[i] <- rnorm(1, candidates$CompX[i], candidates$CompSD[i])
    candidates$BaseDrw[i] <- rnorm(1, candidates$BaseX[i], candidates$BaseSD[i])
    candidates$ActDrw[i] <- rnorm(1, candidates$ActX[i], candidates$ActSD[i])
    candidates$RiskDrw[i] <- rnorm(1, candidates$RiskX[i], candidates$RiskSD[i])
    candidates$ReclassDrw[i] <- rnorm(1, candidates$ReclassX[i], candidates$ReclassSD[i])
    
  }
  candidates$CompDrw <- replace(candidates$CompDrw, candidates$CompDrw>1, 1)
  candidates$CompDrw <- replace(candidates$CompDrw, candidates$CompDrw<0, 0.0001)
  candidates$BaseDrw <- replace(candidates$BaseDrw, candidates$BaseDrw>1, 1)
  candidates$BaseDrw <- replace(candidates$BaseDrw, candidates$BaseDrw<0, 0.0001)
  candidates$ActDrw <- replace(candidates$ActDrw, candidates$ActDrw>1, 1)
  candidates$ActDrw <- replace(candidates$ActDrw, candidates$ActDrw<0, 0.0001)
  candidates$RiskDrw <- replace(candidates$RiskDrw, candidates$RiskDrw>1, 1)
  candidates$RiskDrw <- replace(candidates$RiskDrw, candidates$RiskDrw<0, 0.0001)
  candidates$ReclassDrw <- replace(candidates$ReclassDrw, candidates$ReclassDrw>1, 1)
  candidates$ReclassDrw <- replace(candidates$ReclassDrw, candidates$ReclassDrw<0, 0.0001)
  
  # CALCULATE VALUES FOR INDIVIDUAL OBJECTIVES AND SELECT HIGHEST VALUED ACTIONS WITHIN SCOPE OF BUDGET
  
  # OBJECTIVE 1: Minimize 5-Year Extinction Risk
  # Minimize (Risk*Completion)
  candidates$RiskVal <- candidates$RiskDrw*candidates$CompDrw
  # Create new dataframe to receive data sorted by risk (high to low)
  # Then calculate cumulative cost of actions for this sorting
  risk_only <- candidates[order(-candidates$RiskVal),]
  risk_only$CumSum <- cumsum(risk_only$FWSCostYr)
  risk_only$RiskSel <- risk_only$CumSum < budget
  
  # OBJECTIVE 2: Maximize Increase in Probability of Persistence
  # Maximize (BaselineEX-ActionEX)*Completion
  candidates$PersistVal <- (candidates$BaseDrw-candidates$ActDrw)*candidates$CompDrw
  
  # Because numbers are drawn from distribution, negative persistence numbers could be 
  # calculated.  Any such negative numbers should be reset to 0 change in persistence.
  # recovery Actions cannot have a negative expected effect on persistence in this model.
  for(n in 1:nrow(candidates)){
    if (candidates$PersistVal[n]<0) {
      candidates$PersistVal[n] <- 0 
    }
  }
  
  # Create new dataframe to receive data sorted by persistence (high to low)
  # Then calculate cumulative cost of actions for this sorting
  persist_only <- candidates[order(-candidates$PersistVal),]
  persist_only$CumSum <- cumsum(persist_only$FWSCostYr)
  persist_only$PersistSel <- persist_only$CumSum < budget
  
  
  # OBJECTIVE 3: Maximize Probability AND Value of State Transition
  # From Unknown to Known, Endangered to Threatened, and from Threated to Recovered
  # Assign weights to each possible transition.  Here per email from Cruz-Burgos (Jan 29 2015) 
  # FWS value transitions to Recovered greatest, followed by Endangered to Threatened, followed by
  # Unknown to Known status.
  wUK = 1
  wET = 1.5
  wTR = 3
  # Maximize (Weighted (Reclass*Completion))
  candidates$ReclassVal <- candidates$ReclassDrw*candidates$CompDrw
  
  for(j in 1:nrow(candidates)){
    if (candidates$Knowledge[j]=="U") {
      candidates$ReclassWVal[j] <- wUK*candidates$ReclassVal[j]  
    } else if (candidates$Knowledge[j]=="K" && candidates$Status[j]=="E") {
      candidates$ReclassWVal[j] <- wET*candidates$ReclassVal[j]  
    } else {
      candidates$ReclassWVal[j] <- wTR*candidates$ReclassVal[j]
    }
  }
  
  # Create new dataframe to receive data sorted by reclassification (high to low)
  # Then calculate cumulative cost of actions for this sorting
  reclass_only <- candidates[order(-candidates$ReclassWVal),]
  reclass_only$CumSum <- cumsum(reclass_only$FWSCostYr)
  reclass_only$ReclassSel <- reclass_only$CumSum < budget

  # OBJECTIVE 4: Balance Objectives 1-3
  # The balanced objective is the sum of all weighted values 
  wRisk <- 1.5
  wPersist <- 1.0
  wReclass <- 2.0
  candidates$BalanceVal <- ((wRisk*candidates$RiskVal) + (wPersist*candidates$PersistVal) + (wReclass*candidates$ReclassVal))/
    (wRisk + wPersist + wReclass)
  
  # Create new dataframe to receive data sorted by reclassification (high to low)
  # Then calculate cumulative cost of actions for this sorting
  balance <- candidates[order(-candidates$BalanceVal),]
  balance$CumSum <- cumsum(balance$FWSCostYr)
  balance$BalanceSel <- balance$CumSum < budget
  
  # ORGANIZE TABULAR OBJECTIVE RESULTS
  # Create a clean table of results for reporting and 
  rkeeps <- c("Item", "Species", "Action", "Status", "Knowledge", "Cost", "Time", "Benefits", 
              "Leverage", "Factors", "RiskVal","RiskSel")
  risk_report <- risk_only[,rkeeps,drop=FALSE]
  pkeeps <- c("Item", "PersistVal","PersistSel")
  persist_report <- persist_only[,pkeeps,drop=FALSE]
  ckeeps <- c("Item", "ReclassVal","ReclassSel")
  reclass_report <- reclass_only[,ckeeps,drop=FALSE]
  bkeeps <- c("Item", "BalanceVal","BalanceSel")
  balance_report <- balance[,bkeeps,drop=FALSE]
    
  all_values <-merge(risk_report, persist_report, by.x="Item", by.y="Item")
  all_values <-merge(all_values, reclass_report, by.x="Item", by.y="Item")
  all_values <-merge(all_values, balance_report, by.x="Item", by.y="Item") 

  riskresults[,s+3] <- all_values$RiskSel
  persistresults[,s+3] <- all_values$PersistSel
  reclassresults[,s+3] <- all_values$ReclassSel
  balanceresults[,s+3] <- all_values$BalanceSel
  
} #End of the simulations for-loop

###############################################################

# RESULTS SUMMARY 1: Count how many time individual actions were selected within decision sets

# Convert TRUE/FALSE to 1/0
riskresults[,4:ncol(riskresults)] <- riskresults[,4:ncol(riskresults)]*1
persistresults[,4:ncol(persistresults)] <- persistresults[,4:ncol(persistresults)]*1
reclassresults[,4:ncol(reclassresults)] <- reclassresults[,4:ncol(reclassresults)]*1
balanceresults[,4:ncol(balanceresults)] <- balanceresults[,4:ncol(balanceresults)]*1

# Add column for how many times each action selected
for (x in 1:nrow(candidates)){
 riskresults$RiskCnt[x] <- sum(riskresults[x,4:sims+3])
 persistresults$PersistCnt[x] <- sum(persistresults[x,4:sims+3])
 reclassresults$ReclassCnt[x] <- sum(reclassresults[x,4:sims+3])
 balanceresults$BalanceCnt[x] <- sum(balanceresults[x,4:sims+3])
}

finalcounts = riskresults[,1:3]
finalcounts$RiskCnt = riskresults$RiskCnt
finalcounts$PersistCnt = persistresults$PersistCnt
finalcounts$ReclassCnt = reclassresults$ReclassCnt
finalcounts$BalanceCnt = balanceresults$BalanceCnt

write.table(finalcounts, "ActionCounts.csv", sep=",", row.names=FALSE)

# RESULTS SUMMARY 2: Count how many time a set of actions was selected as a unit.

risk_set <- riskresults
persist_set <- persistresults
reclass_set <- reclassresults
balance_set <- balanceresults

for (y in 1:nrow(candidates)){
  risk_set[y,4:(sims+4)] <- risk_set[y,4:(sims+4)]*risk_set[y,1]
  persist_set[y,4:(sims+4)] <- persist_set[y,4:(sims+4)]*persist_set[y,1]
  reclass_set[y,4:(sims+4)] <- reclass_set[y,4:(sims+4)]*reclass_set[y,1]
  balance_set[y,4:(sims+4)] <- balance_set[y,4:(sims+4)]*balance_set[y,1]  
} # loop applies the draw number to each selection of a candidate action in the draw history
  # (i.e., a four draw history of 0,1,0,1 for action 7 becomes 0,7,0,7)

risk_draws = data.frame(matrix(NA, nrow=sims, ncol=2)) # a table that will have the set number and the set string
colnames(risk_draws) <- c("DrawNumber","DrawSet") 
risk_draws[,1] = as.factor(c(1:sims)) # number of each simulation set of draws in the history for tabular display

persist_draws = data.frame(matrix(NA, nrow=sims, ncol=2))
colnames(persist_draws) <- c("DrawNumber","DrawSet") 
persist_draws[,1] = as.factor(c(1:sims))

reclass_draws = data.frame(matrix(NA, nrow=sims, ncol=2))
colnames(reclass_draws) <- c("DrawNumber","DrawSet") 
reclass_draws[,1] = as.factor(c(1:sims))

balance_draws = data.frame(matrix(NA, nrow=sims, ncol=2))
colnames(balance_draws) <- c("DrawNumber","DrawSet") 
balance_draws[,1] = as.factor(c(1:sims))

for (z in 4:(sims+3)){
  risk_temp <- paste(risk_set[,z],collapse = ",") # collapses each column of draws into a single character string
  risk_a <- as.character(risk_temp)
  risk_b <- strsplit(risk_a, "[^[:digit:]]")
  risk_c <- as.numeric(unlist(risk_b))
  risk_d <- unique(risk_c[!is.na(risk_c)])
  if (risk_d[1]==0){
    risk_d = risk_d[2:length(risk_d)]
  }
    risk_draws[z-3,2] <- paste(risk_d,collapse = ",")
  
  persist_temp <- paste(persist_set[,z],collapse = ",") # collapses each column of draws into a single character string
  persist_a <- as.character(persist_temp)
  persist_b <- strsplit(persist_a, "[^[:digit:]]")
  persist_c <- as.numeric(unlist(persist_b))
  persist_d <- unique(persist_c[!is.na(persist_c)])
  if (persist_d[1]==0){
    persist_d = persist_d[2:length(persist_d)]
  }
  persist_draws[z-3,2] <- paste(persist_d,collapse = ",")
  
  reclass_temp <- paste(reclass_set[,z],collapse = ",") # collapses each column of draws into a single character string
  reclass_a <- as.character(reclass_temp)
  reclass_b <- strsplit(reclass_a, "[^[:digit:]]")
  reclass_c <- as.numeric(unlist(reclass_b))
  reclass_d <- unique(reclass_c[!is.na(reclass_c)])
  if (reclass_d[1]==0){
    reclass_d = reclass_d[2:length(reclass_d)]
  }
  reclass_draws[z-3,2] <- paste(reclass_d,collapse = ",")
  
  balance_temp <- paste(balance_set[,z],collapse = ",") # collapses each column of draws into a single character string
  balance_a <- as.character(balance_temp)
  balance_b <- strsplit(balance_a, "[^[:digit:]]")
  balance_c <- as.numeric(unlist(balance_b))
  balance_d <- unique(balance_c[!is.na(balance_c)])
  if (balance_d[1]==0){
    balance_d = balance_d[2:length(balance_d)]
  }
  balance_draws[z-3,2] <- paste(balance_d,collapse = ",")
  
  }

# Find duplicate sets and calculate frequency
risk_series = as.data.frame(table(risk_draws$DrawSet)) 
persist_series = as.data.frame(table(persist_draws$DrawSet))
reclass_series = as.data.frame(table(reclass_draws$DrawSet))
balance_series = as.data.frame(table(balance_draws$DrawSet))

# Create a table that will have the action item numbers as string sets and percent of total draws for each set
risk_out = data.frame(matrix(NA, nrow=(nrow(risk_series)), ncol=2))
colnames(risk_out) <- c("DecisionSet", "PercentSimulations") 
persist_out = data.frame(matrix(NA, nrow=(nrow(persist_series)), ncol=2))
colnames(persist_out) <- c("DecisionSet", "PercentSimulations") 
reclass_out = data.frame(matrix(NA, nrow=(nrow(reclass_series)), ncol=2)) 
colnames(reclass_out) <- c("DecisionSet", "PercentSimulations") 
balance_out = data.frame(matrix(NA, nrow=(nrow(balance_series)), ncol=2)) 
colnames(balance_out) <- c("DecisionSet", "PercentSimulations") 

# Lookup and match the numbers from the sets with the item numbers to grab the action description
for (i in 1:nrow(risk_series)){
  numacts = as.numeric(strsplit(as.character(risk_series[i,1]), ",")[[1]]) # turn the character strings into sets of numbers
  txtacts = risk_set$Action[match(numacts, risk_set$Item)] # match numbers in sets to Item numbers to find Actions
  lstacts = paste0("(", numacts, ") ", txtacts) # put the matching set of actions into a number list
  risk_out[i,1] <- paste(lstacts,collapse = ", ")
  risk_out[i,2] <- round((risk_series[i,2]/sum(risk_series[,2]))*100, digits=0)
}
risk_ordout <- risk_out[order(-risk_out$PercentSimulations),]
write.table(risk_ordout, "RiskObjectiveSets.csv", sep=",", row.names=FALSE)

for (i in 1:nrow(persist_series)){
  numacts = as.numeric(strsplit(as.character(persist_series[i,1]), ",")[[1]]) # turn the character strings into sets of numbers
  txtacts = persist_set$Action[match(numacts, persist_set$Item)] # match numbers in sets to Item numbers to find Actions
  lstacts = paste0("(", numacts, ") ", txtacts) # put the matching set of actions into a number list
  persist_out[i,1] <- paste(lstacts,collapse = ", ")
  persist_out[i,2] <- round((persist_series[i,2]/sum(persist_series[,2]))*100, digits=0)
}
persist_ordout <- persist_out[order(-persist_out$PercentSimulations),]
write.table(persist_ordout, "PersistObjectiveSets.csv", sep=",", row.names=FALSE)

for (i in 1:nrow(reclass_series)){
  numacts = as.numeric(strsplit(as.character(reclass_series[i,1]), ",")[[1]]) # turn the character strings into sets of numbers
  txtacts = reclass_set$Action[match(numacts, reclass_set$Item)] # match numbers in sets to Item numbers to find Actions
  lstacts = paste0("(", numacts, ") ", txtacts) # put the matching set of actions into a number list
  reclass_out[i,1] <- paste(lstacts,collapse = ", ")
  reclass_out[i,2] <- round((reclass_series[i,2]/sum(reclass_series[,2]))*100, digits=0)
}
reclass_ordout <- reclass_out[order(-reclass_out$PercentSimulations),]
write.table(reclass_ordout, "ReclassObjectivesSets.csv", sep=",", row.names=FALSE)

for (i in 1:nrow(balance_series)){
  numacts = as.numeric(strsplit(as.character(balance_series[i,1]), ",")[[1]]) # turn the character strings into sets of numbers
  txtacts = balance_set$Action[match(numacts, balance_set$Item)] # match numbers in sets to Item numbers to find Actions
  lstacts = paste0("(", numacts, ") ", txtacts) # put the matching set of actions into a number list
  balance_out[i,1] <- paste(lstacts,collapse = ", ")
  balance_out[i,2] <- round((balance_series[i,2]/sum(balance_series[,2]))*100, digits=0)
}
balance_ordout <- balance_out[order(-balance_out$PercentSimulations),]
write.table(balance_ordout, "WeightedObjectivesSets.csv", sep=",", row.names=FALSE)